from typing import Any, Literal, Protocol, overload

from asgi_user_agents.datastructures import UADetails
from asgi_user_agents.types import Scope

try:
    from starlette.requests import Request
except ImportError:  # pragma: no cover
    pass  # Starlette not installed.
else:

    class UAScopeProtocol(Protocol):
        @overload
        def __getitem__(self, key: Literal["ua"]) -> UADetails: ...  # pragma: no cover

        @overload
        def __getitem__(self, key: str) -> Any: ...  # pragma: no cover

    class UAScope(UAScopeProtocol, Scope):
        pass

    class UARequest(Request):
        """
        User-Agent request object with `scope` attribute of type `UAScope`.

        !!! note
            Use this class to make code editors type-check `request.scope["ua"]`.
        """

        scope: UAScope



import httpx
import parametrize_from_file as pff
import pytest
from fastapi.applications import FastAPI
from starlette.middleware import Middleware
from starlette.responses import JSONResponse, Response

from asgi_user_agents import UADetails, UAMiddleware
from asgi_user_agents import UARequest as Request

app = FastAPI(middleware=[Middleware(UAMiddleware)])


@app.get("/")
async def index(request: Request) -> Response:
    ua = request.scope["ua"]
    assert isinstance(ua, UADetails)
    data = {
        "ua_string": ua.ua_string,
        "os": ua.os,
        "os.family": ua.os.family,
        "os.version": ua.os.version,
        "os.version_string": ua.os.version_string,
        "browser": ua.browser,
        "browser.family": ua.ua.browser.family,
        "browser.version": ua.ua.browser.version,
        "browser.version_string": ua.ua.browser.version_string,
        "device": ua.device,
        "device.family": ua.device.family,
        "device.brand": ua.device.brand,
        "device.model": ua.device.model,
        "is_provided": ua.is_provided,
        "is_tablet": ua.is_tablet,
        "is_mobile": ua.is_mobile,
        "is_touch_capable": ua.is_touch_capable,
        "is_pc": ua.is_pc,
        "is_bot": ua.is_bot,
        "is_email_client": ua.is_email_client,
    }
    return JSONResponse(data)


@pytest.mark.asyncio
@pff.parametrize(path="assets/test_middleware.json")
async def test_user_agent_data(ua_string: str, response_data: dict) -> None:
    async with httpx.AsyncClient(app=app) as client:
        response = await client.get(
            "http://testserver/", headers={"User-Agent": ua_string}
        )
        data = response.json()
        assert data["ua_string"] == response_data["ua_string"]
        assert data["os"] == response_data["os"]
        assert data["os.family"] == response_data["os.family"]
        assert data["os.version"] == response_data["os.version"]
        assert data["os.version_string"] == response_data["os.version_string"]
        assert data["browser"] == response_data["browser"]
        assert data["browser.family"] == response_data["browser.family"]
        assert data["browser.version"] == response_data["browser.version"]
        assert data["browser.version_string"] == response_data["browser.version_string"]
        assert data["device"] == response_data["device"]
        assert data["device.family"] == response_data["device.family"]
        assert data["device.brand"] == response_data["device.brand"]
        assert data["device.model"] == response_data["device.model"]
        assert data["is_provided"] is response_data["is_provided"]
        assert data["is_tablet"] is response_data["is_tablet"]
        assert data["is_mobile"] is response_data["is_mobile"]
        assert data["is_touch_capable"] is response_data["is_touch_capable"]
        assert data["is_pc"] is response_data["is_pc"]
        assert data["is_bot"] is response_data["is_bot"]
        assert data["is_email_client"] is response_data["is_email_client"]



from contextlib import suppress

from asgi_user_agents.datastructures import UADetails as UADetails
from asgi_user_agents.middleware import UAMiddleware as UAMiddleware

with suppress(ImportError):
    from asgi_user_agents.requests import UARequest as UARequest

__version__ = "0.1.0"

__all__ = [
    "__version__",
    "UAMiddleware",
    "UADetails",
    "UARequest",
]



from __future__ import annotations

from typing import Optional
from urllib.parse import unquote

from user_agents.parsers import Browser, Device, OperatingSystem, UserAgent

from asgi_user_agents.types import Scope


class UADetails:
    def __init__(self, scope: Scope) -> None:
        self._scope = scope
        self._ua: UserAgent

    def _get_header(self, name: bytes) -> Optional[str]:
        return _get_header(self._scope, name)

    @property
    def _ua_string(self) -> Optional[str]:
        return self._get_header(b"User-Agent") or ""

    @property
    def ua(self) -> UserAgent:
        if not hasattr(self, "_ua"):
            self._ua = UserAgent(self._ua_string)
        return self._ua

    @property
    def ua_string(self) -> str:
        return self.ua.ua_string

    @property
    def is_provided(self) -> bool:
        return bool(self.ua_string)

    @property
    def os(self) -> OperatingSystem:
        return self.ua.os

    @property
    def browser(self) -> Browser:
        return self.ua.browser

    @property
    def device(self) -> Device:
        return self.ua.device

    @property
    def is_tablet(self) -> bool:
        return self.ua.is_tablet

    @property
    def is_mobile(self) -> bool:
        return self.ua.is_mobile

    @property
    def is_touch_capable(self) -> bool:
        return self.ua.is_touch_capable

    @property
    def is_pc(self) -> bool:
        return self.ua.is_pc

    @property
    def is_bot(self) -> bool:
        return self.ua.is_bot

    @property
    def is_email_client(self) -> bool:
        return self.ua.is_email_client


def _get_header(scope: Scope, key: bytes) -> Optional[str]:
    key = key.lower()
    value: Optional[str] = None
    should_unquote = False

    for k, v in scope["headers"]:
        if k.lower() == key:
            value = v.decode("latin-1")
        if k.lower() == b"%s-uri-autoencoded" % key and v == b"true":
            should_unquote = True

    if value is None:
        return None

    return unquote(value) if should_unquote else value
